import EMDY from "@endeavorance/emdy";
import TOML from "smol-toml";
import YAML from "yaml";
import { loadFileContent } from "./files";
import type { UnknownRecord } from "./types";

export interface MuseEntry<MetaShape = UnknownRecord> {
  _raw: string;
  filePath: string;
  data: Record<string, unknown>;
  meta: MetaShape;
}

function formatEntries(val: unknown): UnknownRecord[] {
  if (Array.isArray(val) && val.every((el) => typeof el === "object")) {
    return val;
  }

  if (typeof val === "object" && val !== null) {
    return [val as UnknownRecord];
  }

  throw new Error(
    `Invalid data format. Entry files must define an object or array of objects, but found "${val}"`,
  );
}

function parseYAMLEntries(text: string): UnknownRecord[] {
  const parsedDocs = YAML.parseAllDocuments(text);

  if (parsedDocs.some((doc) => doc.toJS() === null)) {
    throw new Error("Encountered NULL resource");
  }

  const errors = parsedDocs.flatMap((doc) => doc.errors);

  if (errors.length > 0) {
    throw new Error(
      `Error parsing YAML resource: ${errors.map((e) => e.message).join(", ")}`,
    );
  }

  const collection: UnknownRecord[] = parsedDocs.map((doc) => doc.toJS());
  return collection;
}

function parseJSONEntries(text: string): UnknownRecord[] {
  return formatEntries(JSON.parse(text));
}

function parseTOMLEntries(text: string): UnknownRecord[] {
  return formatEntries(TOML.parse(text));
}

function parseMarkdownEntry(text: string, contentKey: string): UnknownRecord[] {
  return formatEntries(EMDY.parse(text, contentKey));
}

interface ParseMuseFileOptions {
  contentKey?: string;
}

export async function parseMuseFile(
  rawFilePath: string,
  { contentKey = "content" }: ParseMuseFileOptions = {},
): Promise<MuseEntry[]> {
  const { content, filePath, fileType } = await loadFileContent(rawFilePath);

  const partial = {
    _raw: content,
    filePath,
    meta: {},
  };

  if (fileType === "md") {
    return parseMarkdownEntry(content, contentKey).map((data) => ({
      ...partial,
      data,
    }));
  }

  if (fileType === "yaml") {
    return parseYAMLEntries(content).map((data) => ({
      ...partial,
      data,
    }));
  }

  if (fileType === "json") {
    return parseJSONEntries(content).map((data) => ({
      ...partial,
      data,
    }));
  }

  if (fileType === "toml") {
    return parseTOMLEntries(content).map((data) => ({
      ...partial,
      data,
    }));
  }

  throw new Error(`Unsupported file type: ${fileType}`);
}
